'''
findmalware - Version 1.0b (2013-11-06).
Latest version at http://www.itsec.se

This is a generic malware scanner used to identify Windows
portable executable (PE) files with "suspicious" attributes. The
script is a modified version of pescanner.py by Michael Ligh and
uses pefile by Ero Carrera. It is designed to parse a large
number of files and to list executables that warrant additional
attention (e.g. Top 10). It is suited for investigations where
malware is suspected but where traditional antivirus software are
unable to find anything. The script uses basic heuristics detailed
in the paper "Pimp my PE - Parsing Malicious and Malformed
Executables" by Sunbelt Software. Please note that the number of
identifiers have been decreased in order to improve performance and
to reduce dependencies.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the named License,
or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

Copyright (C) 2013 by Gustav Nordenskjold, Sweden.
'''

#!/usr/bin/python

import time
import binascii
import string
import os, sys

try:
	import pefile
	import peutils
except ImportError:
	print 'pefile not installed, see http://code.google.com/p/pefile/'
	sys.exit()

# suspicious APIs to alert on 
alerts = [	'OpenProcess', 'VirtualAllocEx', 'WriteProcessMemory',
			'CreateRemoteThread', 'ReadProcessMemory',
			'CreateProcess', 'WinExec', 'ShellExecute',
			'HttpSendRequest', 'InternetReadFile', 'InternetConnect',
			'CreateService', 'StartService']
		  
# legit entry point sections
good_ep_sections = ['.text', '.code', 'CODE', 'INIT', 'PAGE']

def usage():
	print 'Usage: python ' + sys.argv[0] + ' path [-t int] [--list] [--top] [--help]'
	print 'path\t\tPath to search for files (file or directory).'
	print '-t integer\tThreshold for amount of suspicious attributes (default is 4)'
	print '--list\t\tList of all binary files with suspicious attributes above the threshold (-t)'
	print '--top\t\tList the 10 binary files with the most suspicious attributes'
	print '--help\t\tThis help text.'
	print ''

def convert_char(char):
	if char in string.ascii_letters or \
		char in string.digits or \
		char in string.punctuation or \
		char in string.whitespace:
		return char
	else:
		return r'\x%02x' % ord(char)

def convert_to_printable(s):
	return ''.join([convert_char(c) for c in s])

class MalwareScanner:
	def __init__(self, peid_sigs=None):

		# initialize PEiD signatures if provided 
		if peid_sigs:
			self.sigs = peutils.SignatureDatabase(peid_sigs)
		else:
			self.sigs = None
		
	def check_ep_section(self, pe):
		""" Determine if a PE's entry point is suspicious """
		name = ''
		ep = pe.OPTIONAL_HEADER.AddressOfEntryPoint
		pos = 0
		for sec in pe.sections:
			if (ep >= sec.VirtualAddress) and \
			   (ep < (sec.VirtualAddress + sec.Misc_VirtualSize)):
				name = sec.Name.replace('\x00', '')
				break
			else: 
				pos += 1
		return (ep, name, pos)

	def check_verinfo(self, pe):
		""" Determine the version info in a PE file """
		ret = []
		
		if hasattr(pe, 'VS_VERSIONINFO'):
			if hasattr(pe, 'FileInfo'):
				for entry in pe.FileInfo:
					if hasattr(entry, 'StringTable'):
						for st_entry in entry.StringTable:
							for str_entry in st_entry.entries.items():
								ret.append(str_entry[0] + ': ' + str_entry[1] )
					elif hasattr(entry, 'Var'):
						for var_entry in entry.Var:
							if hasattr(var_entry, 'entry'):
								ret.append(var_entry.entry.keys()[0] + ': ' + var_entry.entry.values()[0])
		return '\n'.join(ret)

	def check_tls(self, pe):
		callbacks = []
		if (hasattr(pe, 'DIRECTORY_ENTRY_TLS') and \
					pe.DIRECTORY_ENTRY_TLS and \
					pe.DIRECTORY_ENTRY_TLS.struct and \
					pe.DIRECTORY_ENTRY_TLS.struct.AddressOfCallBacks):
			callback_array_rva = pe.DIRECTORY_ENTRY_TLS.struct.AddressOfCallBacks - pe.OPTIONAL_HEADER.ImageBase 
			idx = 0
			while True:
				func = pe.get_dword_from_data(pe.get_data(callback_array_rva + 4 * idx, 4), 0)
				if func == 0: 
					break
				callbacks.append(func)
				idx += 1
		return callbacks
	
	def check_imports(self, pe):
		ret = []
		if not hasattr(pe, 'DIRECTORY_ENTRY_IMPORT'):
			return ret
		for lib in pe.DIRECTORY_ENTRY_IMPORT:
			for imp in lib.imports:
				if (imp.name != None) and (imp.name != ""):
					for alert in alerts:
						if imp.name.startswith(alert):
							ret.append(imp.name)
		return ret

	def check_packers(self, pe):
		packers = []
		if self.sigs:
			matches = self.sigs.match(pe, ep_only = True)
			if matches != None:
				for match in matches:
					packers.append(match)
		return packers

	def check_rsrc(self, pe):
		ret = {}
		binary = False
		if hasattr(pe, 'DIRECTORY_ENTRY_RESOURCE'):
			i = 0
			for resource_type in pe.DIRECTORY_ENTRY_RESOURCE.entries:
				if resource_type.name is not None:
					name = "%s" % resource_type.name
				else:
					name = "%s" % pefile.RESOURCE_TYPE.get(resource_type.struct.Id)
				if name == None:
					name = "%d" % resource_type.struct.Id
				if hasattr(resource_type, 'directory'):
					for resource_id in resource_type.directory.entries:
						if hasattr(resource_id, 'directory'):
							for resource_lang in resource_id.directory.entries:
								data = pe.get_data(resource_lang.data.struct.OffsetToData, resource_lang.data.struct.Size)

								lang = pefile.LANG.get(resource_lang.data.lang, '*unknown*')
								sublang = pefile.get_sublang_name_for_lang( resource_lang.data.lang, resource_lang.data.sublang )
								if data.encode('hex').lower().startswith('4d5a'):
									ret[i] = (name, resource_lang.data.struct.OffsetToData, resource_lang.data.struct.Size, lang, sublang)
									i += 1
									binary = True
								elif data.encode('hex').lower().startswith('5a4d'):
									ret[i] = (name, resource_lang.data.struct.OffsetToData, resource_lang.data.struct.Size, lang, sublang)
									i += 1
									binary = True
								else:
									ret[i] = (name, resource_lang.data.struct.OffsetToData, resource_lang.data.struct.Size, lang, sublang)
									i += 1
		return binary, ret
	
	def start_scan( self, file ):
		count = 0
		output = []

		try:
			FILE = open(file, "rb")
			data = FILE.read()
			FILE.close()
		except:
			output.append("Read error (%s)" % file)
			return -1, output

		if data == None or len(data) == 0:
			output.append("Read error. File appears to be empty (%s)" % file)
			return -1, output
			
		try:
			pe = pefile.PE(data=data, fast_load=True)
			pe.parse_data_directories( directories=[ 
				pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_IMPORT'],
				pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_EXPORT'],
				pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_TLS'],
				pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_RESOURCE']])
		except:
			output.append("Parsing error. Not a PE-file.  (%s)" % file)
			return -1, output
		
		output.append("" )
		output.append("========================================================================")

		#FILE INFO
		output.append("File:	%s" % file)
		output.append("Size:	%d bytes" % len(data))
		output.append("Date:	%s UTC" % time.asctime(time.gmtime(pe.FILE_HEADER.TimeDateStamp)) )
		output.append("" )
		
		#Signature			
		verinfo = self.check_verinfo(pe)
		if len(verinfo):
			output.append("Version information:")
			output.append( verinfo )
			output.append("")

		output.append("------------------------------------------------------------------------")
		output.append("")
		
		try:
			that_year = time.gmtime( pe.FILE_HEADER.TimeDateStamp )[0]
			this_year = time.gmtime(time.time())[0]
			if that_year < 2000 or that_year > this_year:
				count += 1
				output.append("Invalid timestamp:")
				output.append("%s UTC" % time.asctime(time.gmtime(pe.FILE_HEADER.TimeDateStamp)) )
				output.append("")
		except:
			output.append("Invalid timestamp:")
			output.append("%s UTC" % time.asctime(time.gmtime(pe.FILE_HEADER.TimeDateStamp)) )
			output.append("")
			count += 1
		
		# EP section
		(ep, name, pos) = self.check_ep_section(pe)
		if (name not in good_ep_sections) or pos == len(pe.sections):
			count += 1
			output.append("Suspicious entry point section(s):")
			output.append( "EP: %s %s %d/%d" % (hex(ep+pe.OPTIONAL_HEADER.ImageBase), name, pos, len(pe.sections)) )
			output.append("")
			
		#CRC checksum
		if (pe.OPTIONAL_HEADER.CheckSum):
			crc_claimed = pe.OPTIONAL_HEADER.CheckSum
			crc_actual  = pe.generate_checksum()
			if crc_actual != crc_claimed:
				count += 1
				output.append("Suspicious checksums:")
				output.append("Claimed CRC: 0x%x" % (crc_claimed) )
				output.append("Actual CRC: 0x%x" % (crc_actual) )
				output.append("")
			
		#Packer
		packers = self.check_packers(pe)
		if len(packers):
			count +=1
			output.append("Packer signature identified:")
			output.append("Packers: %s" % ','.join(packers))
			output.append("")
	 
		#TLS callbacks = Suspicious
		callbacks = self.check_tls(pe)
		if len(callbacks):
			count += 1
			output.append("Suspicious TLS entries:")
			for cb in callbacks:
				output.append("	0x%x" % cb)
			output.append("")
						
		#Suspicious IAT entries
		imports = self.check_imports(pe)
		if len(imports):
			count += 1
			output.append("Suspicious IAT entries:")
			for imp in imports:
				output.append(imp)
			output.append("")
		
		#Check resources
		binary, rsrc = self.check_rsrc(pe)
		if binary and len(rsrc):
			count += 1
			output.append("Binary resources found:")
			output.append("%-18s %-8s %-8s %-12s %-24s" % ("Name", "RVA", "Size", "Lang", "Sublang"))
			for r in rsrc:
				(name,rva,size,lang,sublang) = rsrc[r]
				output.append("%-18s %-8s %-8s %-12s %-24s" % (name, hex(rva), hex(size), lang, sublang)) 
			output.append("")
			  
		#Suspicious Sections
		sections = False
		for sec in pe.sections:
			if sec.SizeOfRawData == 0 or (sec.get_entropy() > 0 and sec.get_entropy() < 1) or sec.get_entropy() > 7:
				if sections == False:
					count += 1
					output.append("Suspicious sections found:")
					output.append("%-10s %-12s %-12s %-12s %-12s" % ("Name", "VirtAddr", "VirtSize", "RawSize", "Entropy"))
					sections = True
				
				output.append("%-10s %-12s %-12s %-12s %-12f" % (
					''.join([c for c in sec.Name if c in string.printable]), 
					hex(sec.VirtualAddress), hex(sec.Misc_VirtualSize), 
					hex(sec.SizeOfRawData), sec.get_entropy()) )
		if sections:
			output.append("")
		
		return count, output
			
if __name__ == '__main__':

	print 'findmalware - find potential malware based on suspicious PE header attributes.'
	print 'Version 1.0b (2013-11-06). Latest version at http://www.itsec.se'
	print ''
	
	if len(sys.argv) < 2:
		usage()
		sys.exit(0)
	
	paths = []
	TYPE = "Standard"
	THRESHOLD = 4 		# This value should exclude most false positives (and possibly some malware as well)
	THREATEXPERT = False
	
	paths.append( sys.argv[1].strip('"') )  #strip is used to handle trailing backslashes
			
	for i in range( len(sys.argv) ):
		if sys.argv[i] == '-t':
			try:
				THRESHOLD = int(sys.argv[i+1])
			except:
				print 'Unable to parse threshold (-t).'
				sys.exit(0)
		elif sys.argv[i] == '--list':
			TYPE = "List"
			print 'Count : Path'
		elif sys.argv[i] == '--top':
			TYPE = "Top10"
			print 'Count : Path'
		elif sys.argv[i] == '--help':
			usage()
			sys.exit(0)
	
	'''if not paths:
		if( os.name == 'nt' ):
			for i in range(ord('a'), ord('z')+1):
				drive = chr(i)
				if(exists(drive +":\\")):
					paths.append(drive +":\\")
		else:
			paths.append('/')
	'''	
	
	scanner = MalwareScanner('UserDB.TXT')
	files  = []
	toplist = []
	
	
	for path in paths:
		if os.path.isdir( path ):
			for root, dirs, files in os.walk( path ):
				for file in files:
				
					count, output = scanner.start_scan( os.path.join(root, file) )
				
					if TYPE == "Standard" and count >= THRESHOLD:
						print '\n'.join([convert_to_printable(s) for s in output])
					elif TYPE == "List" and count >= THRESHOLD:
						print str(count) + " : " + os.path.join(root, file)
					elif TYPE == "Top10" and count >= THRESHOLD:
						toplist.append( (count, os.path.join(root, file)) )
				
			if TYPE == "Top10" and toplist:
				sorted_list = sorted(toplist, key=lambda c: c[0], reverse=True)
				for i in sorted_list[:10]:
					print str(i[0]) + " : " + str(i[1]),
					print ''
					
		elif os.path.isfile( path ):

			count, output = scanner.start_scan( path )
			print '\n'.join([convert_to_printable(s) for s in output])

		else:
			print "You must supply a file or directory! (" + path + ")"
			sys.exit()
